/*
 *  controller.c - Generic controller-functions.
 *
 *  (C) 1994, 1995 Mikael Nordqvist (d91mn@efd.lth.se, mech@df.lth.se)
 */

#include <unistd.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <errno.h>
#include <string.h>

#include "mod.h"
#include "lists.h"
#include "message.h"


void init_controller(void)
{
    total_time=0;
    set_playing(0);

    if(GUI_mode == GUI_SIMPLE)
	simple_init_controller();
#ifdef COMPILE_GUI_NC
    if(GUI_mode == GUI_NC)
	nc_init_controller();
#endif
#ifdef COMPILE_GUI_TK
    if(GUI_mode == GUI_TK)
	tk_init_controller();
#endif
    if(GUI_mode == GUI_NONE) {
	L_order_reset(PL);
	load_next(1, 1, 1);
    }
}


/* Returns true if the module could be loaded */

int change_module(char *filename, char *options)
{
    char buf[80];
    int nr;
    
    debug("Entering load_module()");

    /* Release resources used by the previous module */
    if(loaded) {
	debug("Ordering player to exit...\n");
	command_player(MESSAGE_EXIT, 42);
	
	debug("Drainpipes(0)...\n");
	drain_pipes(0);
	debug("Drainpipes(0) Done.\n");
	do {
	    debug("Waitpid(pid_player)...\n");
	    if((nr=waitpid(pid_player, 0, 0)) == -1 && errno != EINTR)
		error("waitpid() failed in main.\n");
	}
	while(nr != pid_player);
	debug("Drainpipes(1)...\n");
	drain_pipes(1);
	debug("Drainpipes(1) done.\n");
	free_module();
	debug("Module free'd.\n");
	loaded=0;
    }
    
    pre_load_update_screen();
    
    /* Make sure sequencer is clean */
    debug("cleanup_sound()...\n");
    cleanup_sound();
    debug("init_sound()...\n");
    init_sound();
    debug("init_sound() Done, loading next module...\n\n");
    
    opt=default_opt;
    strcpy(M.options_str, options);
    process_options_str(options, &opt);

    maxtime_done=0;

    print_filename(path_basename(filename, 0));
    if(!load_module(filename)) {
	if(quit != QUIT_SIGNAL) {
	    print_status("Loading failed");
	    sleep(1);
	    print_filename("");
	    error_load_update_screen();
	    set_playing(0);
	    free_module();
	    loaded=0;
	    return 0;
	}
	else {
	    return 0; /* Leave if we got a Ctrl-C */
	}  
    }
    loaded=LOADED_VOID; /* Set loaded to VOID for now, and let interface
			 * change to FILE/LIST later.
			 */
    print_songinfo(0);
    
    switch(pid_player=fork()) {
      case -1:
	error("Fork failed.\n");
	break;
	
      case 0:  /* Child - player */
	install_ignore_suspend_handler();
	debug("Player calling play_module()\n");
	play_module();
	cleanup_sound();
	exit(0);
	break;
	
      default: /* Parent */
	debug("Parent continuing execution after fork().\n");
    }
    
    starting_time=restart_time=time(0);
    set_playing(1);

    last_speed=6; /* Defaults if a key is hit b4 1st line is played */
    last_tempo=125;

    post_load_update_screen();

    if(GUI_mode != GUI_TK && !quiet)
	read(STDIN_FILENO, buf, 80); /* Flush input */
    return 1;                    /* Loading successful */
}


void end_of_song(void)
{
    if(GUI_mode == GUI_SIMPLE)
	simple_end_of_song();
#ifdef COMPILE_GUI_NC
    if(GUI_mode == GUI_NC)
	nc_end_of_song();
#endif
#ifdef COMPILE_GUI_TK
    if(GUI_mode == GUI_TK)
	tk_end_of_song();
#endif
    if(GUI_mode == GUI_NONE)
	load_next(1, 1, 1);
}


void set_playing(int a)
{
    if(a) {
	if(loaded == LOADED_LIST)
	    print_status("Playing list...");
	else if(loaded == LOADED_FILE)
	    print_status("Playing directory...");
	else
	    print_status("Playing...");
	playing=1;
    }
    else {
	print_status("Ready...");
	playing=0;
    }
}

int change_dir(char *dir)
{
    int ret=0;

    if(!chdir(dir)) {
	getcwd(workdir, PATH_MAX); /* Change working directory */
	if(loaded == LOADED_FILE)
	    loaded=LOADED_VOID;
	ret=1;
    }

    return ret;
}


/* load_next/prev() attempts to load the next/previous module in the playlist.
 *
 *  Returns 0 if loading failed. Nonzero means loading was successful. Further-
 *  more returnvalues < 0 indicates the last/first module was just loaded.
 *
 * exit_last: If true, player will exit after the last module.
 *
 * exit_last_err: If true the player will only exit after the the last module
 *                if the last song failed to load.
 *
 * persistent: Attempts will be me made to load next module if the current
 *             wouldn't load. If the last module won't load exit_after_load
 *             decides if we quit or just accept the failure.
 */

int load_next(int exit_last, int exit_last_err, int persistent)
{
    char name[PATH_MAX+1];
    int tmp=0; /* Remove warning */
    
    if(loaded != LOADED_FILE) {
	tmp=load_it(0, exit_last, exit_last_err, persistent);
    }
    else {
	if(loaded_filenr+1 >= nr_dirs && loaded_filenr+1 < nr_files) {
	    ++loaded_filenr;
	    
	    strcpy(name, workdir);
	    strcat(name, "/");
	    strcat(name, get_filename(loaded_filenr));
	    tmp=change_module(name, "");
	    if(tmp) {
		loaded=LOADED_FILE;
		if(loaded_filenr == nr_files -1)
		    tmp=-1;
		else
		    tmp=1;
	    }
	    else
		loaded=0;
	}
    }
    set_playing(playing);    
    return tmp;
}

int load_prev(int exit_last, int exit_last_err, int persistent)
{
    char name[PATH_MAX+1];
    int tmp=0; /* Remove warning */

    if(loaded != LOADED_FILE) {
	return load_it(1, exit_last, exit_last_err, persistent);
    }
    else {
	if(loaded_filenr-1 >= nr_dirs && loaded_filenr-1 < nr_files) {
	    --loaded_filenr;
	    
	    strcpy(name, workdir);
	    strcat(name, "/");
	    strcat(name, get_filename(loaded_filenr));
	    tmp=change_module(name, "");
	    if(tmp) {
		loaded=LOADED_FILE;
		if(loaded_filenr == nr_dirs)
		    tmp=-1;
		else
		    tmp=1;
	    }
	    else
		loaded=0;
	}
    }
    set_playing(playing);    
    return tmp;
}

int load_it(int backward, int exit_last, int exit_last_err, int persistent)
{
    int retry=0, do_exit=0, tmp=0, did_load=0;
    int a;

    if(!L_size(PL))
	return 0;
    
    while(!quit) {
	a=(backward ? L_order_prev(PL) : L_order_next(PL));
	if(!a && opt.replay_forever) {
	    if(backward)
		L_order_set(PL, L_size(PL)-1);
	    else
		L_order_set(PL, 0);
	    a=1;
	}
	if(a) {
	    /* Not a repeat of last song */
	    if((tmp=change_module(PL->current->filename,
				  PL->current->options))) {
		did_load=1;
		break; /* Found one, done. */
	    }
	    if(!persistent)
		break;
	}
	else { /* Last one already played */
	    if(exit_last || (retry && exit_last_err))
		do_exit=1;
	    break;
	}
	retry=1;
    }
    
    /* Make sure the last player-process exits if no suitable song could
     * be loaded.
     */
    if(do_exit && !quit) {
	if(loaded)
	    command_player(MESSAGE_EXIT, 42);
	quit=QUIT_QUIT;
    }
    else {
	if(did_load)
	    loaded=LOADED_LIST;
	set_playing(playing);
	cur_listnr=L_order_name_linearget(PL, M.real_filename);
    }

    if(tmp) {
	tmp=1;
	if(backward) {
	    if(!L_order_get(PL))
		tmp=-1;
	}
	else {
	    if(L_order_get(PL) == L_size(PL)-1)
		tmp=-1;
	}
    }
    return tmp;
}
