/*
 *  lists.c - Handling of play-lists.
 *
 *  (C) 1994, 1995 Mikael Nordqvist (d91mn@efd.lth.se, mech@df.lth.se)
 */

#include <stdlib.h>
#include <string.h>

#include "mod.h"
#include "lists.h"
#include <stdio.h>


/* The following functions are called by the user */

void L_init(struct list_head *l)
{
    memset(l, 0, sizeof(struct list_head));
}

void L_append(struct list_head *l, char *filename, char *options)
{
    L_insert_last(l, L_create_item(filename, options));
}


void L_prepend(struct list_head *l, char *filename, char *options)
{
    L_insert_first(l, L_create_item(filename, options));
}


void L_insert_before_nr(struct list_head *l, int nr, char *filename, char *options)
{
    if(!nr)
	L_prepend(l, filename, options);
    else
	L_insert_middle(l, nr-1, L_create_item(filename, options));
}


void L_insert_after_nr(struct list_head *l, int nr, char *filename, char *options)
{
    if(nr >= l->size-1)
	L_append(l, filename, options);
    else
	L_insert_middle(l, nr, L_create_item(filename, options));
}


void L_remove_first(struct list_head *l)
{
    struct list_item *i;

    if(l->first) {
	i=l->first;
	l->first=i->next;
	if(l->first)
	    l->first->prev=0;
	if(l->last == i)
	    l->last=l->first;
	free(i->filename);
	free(i->options);
	free(i);
    }
    else
	error("Removing from empty list...\n");
    --l->size;
}


void L_remove_last(struct list_head *l)
{
    struct list_item *i;

    if(l->last) {
	i=l->last;
	free(i->filename);
	free(i->options);
	l->last=i->prev;
	if(l->last)
	    l->last->next=0;
	if(l->first == i)
	    l->first=l->last;
	free(i);
    }
    else
	error("Removing from empty list...\n");
    --l->size;
}


void L_remove_nr(struct list_head *l, int nr)
{
    struct list_item *i;

    if(!nr)
	L_remove_first(l);
    else if(nr == l->size-1)
	L_remove_last(l);
    else {
	for(i=l->first; nr; i=i->next, --nr)
	    ;
	i->prev->next=i->next;
	i->next->prev=i->prev;
	free(i->filename);
	free(i->options);
	free(i);
	--l->size;
    }
}


struct list_item *L_get_first(struct list_head *l)
{
    return l->first;
}


int L_size(struct list_head *l)
{
    return l->size;
}


void L_clear(struct list_head *l)
{
    while(L_size(l))
	L_remove_first(l);
}


int L_isatend(struct list_head *l)
{
    if(l->cur_nr == -1)
	return 0;
    return l->cur_nr == l->size-1;
}


int L_isatstart(struct list_head *l)
{
    if(l->cur_nr == -1)
	return 0;
    return !l->cur_nr;
}


int L_isreset(struct list_head *l)
{
    return l->cur_nr == -1;
}


void L_set_options_nr(struct list_head *l, int nr, char *new_options)
{
    struct list_item *i;
    
    for(i=l->first; nr; i=i->next, --nr)
	;
    free(i->options);
    i->options=strdup(new_options);
}

char *L_get_options_nr(struct list_head *l, int nr)
{
    struct list_item *i;
    
    for(i=l->first; nr; i=i->next, --nr)
	;
    return i->options;
}

char *L_linear_get_filename_nr(struct list_head *l, int nr)
{
    struct list_item *i;
    
    for(i=l->first; nr; i=i->next, --nr)
	;
    return i->filename;
}

/* 
 * Iterating over the list.
 */

void L_iterate_reset(struct list_head *l)
{
    l->current_iter=l->first;
}


int L_iterate_valid(struct list_head *l)
{
    return l->current_iter != 0;
}


void L_iterate_forward(struct list_head *l)
{
    l->current_iter=l->current_iter->next;
}


void L_iterate_backward(struct list_head *l)
{
    l->current_iter=l->current_iter->prev;
}

/*
 *  Iterating using random/linear order (for play-sequence).
 */

/* Prepare a list. Next call to L_order_next() will get the first module */

void L_order_reset(struct list_head *l)
{
    struct list_item *i;
    int nr, k;
    
    if(default_opt.random_mode) {
	/* Random */
	for(i=l->first; i; i=i->next)
	    i->seq_nr=-1;
	
	for(k=0; k < l->size; ++k) {
	    nr=rand()%(l->size-k);
	    for(i=l->first; nr >= 0; i=i->next)
		if(i->seq_nr < 0)
                    --nr;
	    if(i)
		i->prev->seq_nr=k;
	    else
		l->last->seq_nr=k;
	}
    }
    else {
	/* Linear */
	for(nr=0, i=l->first; i; i=i->next)
	    i->seq_nr=nr++;
    }
    
    l->cur_nr=-1;
}

/* "Moves" the specified module first on the "randomized list", by swapping
 *  positions of 'nr' and the first module. nr is in linear order.
 */

void L_order_linear_swapfirst(struct list_head *l, int nr)
{
    struct list_item *i, *zeroi=0, *nri=0;
    int curnr;

    /* Find the first module */
    for(i=l->first; i; i=i->next)
	if(!i->seq_nr) {
	    zeroi=i;
	    break;
	}
    
    for(curnr=0, i=l->first; i; i=i->next, ++curnr)
	if(curnr == nr) {
	    nri=i;
	    break;
	}
    
    if(zeroi && nri) {
	zeroi->seq_nr=nri->seq_nr;
	nri->seq_nr=0;
    }
}

/* Set current module. nr is the linear order of the module */

void L_order_linearset(struct list_head *l, int nr)
{
    struct list_item *i;
    
    for(i=l->first; nr; i=i->next, --nr)
	;
    l->current=i;
    l->cur_nr=i->seq_nr;
}


void L_order_set(struct list_head *l, int nr)
{
    struct list_item *i;
    
    for(i=l->first; i && i->seq_nr != nr; i=i->next)
	;
    l->current=i;
    l->cur_nr=i->seq_nr;
}


int L_order_get(struct list_head *l)
{
    return l->cur_nr;
}


int L_order_name_linearget(struct list_head *l, char *s)
{
    struct list_item *i;
    int nr;
    
    for(nr=0, i=l->first; i; i=i->next, ++nr)
	if(!strcmp(s, i->filename)) {
	    return nr;
	}
    return -1;
}


/* Returns true until l->current is a repeat of the last module */

int L_order_next(struct list_head *l)
{
    struct list_item *i;
    int ret_val=1;
    
    ++l->cur_nr;
    if(l->cur_nr >= l->size) {
	l->cur_nr=l->size-1;
	ret_val=0;
    }
    
    for(i=l->first; i->seq_nr != l->cur_nr; i=i->next)
	;
    l->current=i;
    return ret_val;
}

/* Returns true until l->current is a repeat of the first module */

int L_order_prev(struct list_head *l)
{
    struct list_item *i;
    int ret_val=1;
    
    if(l->cur_nr == -1) /* First time */
	return 0;

    if(--l->cur_nr < 0) {
	l->cur_nr=0;
	ret_val=0;
    }
    
    for(i=l->first; i->seq_nr != l->cur_nr; i=i->next)
	;
    l->current=i;
    return ret_val;
}


#define MOD_MAGIC "# Mod playlist format version 1.\n"

/* Loading and saving of lists. */

int load_playlist(char *filename, struct list_head *l)
{
    char buf[1024], name[1024], *options;
    FILE *fp;
    int bufpos, namepos;
    
    if(!(fp=fopen(filename, "r")))
	return 0;
    
    if(!fgets(buf, 1024, fp) || strcmp(buf, MOD_MAGIC)) {
	fclose(fp);
	return 0;
    }
    
    L_clear(l);
    
    while(fgets(buf, 1024, fp)) {
	/* Skip blank lines / lines starting with ws / comments */
	if(buf[0] == '\n' || buf[0] == ' ' || buf[0] == '\t' || buf[0] == '#')
	    continue;
	buf[1023]=0;
	for(namepos=bufpos=0; bufpos < 1023;) { /* not 1024 bcause '\' */
	    if(buf[bufpos] == ' ' || buf[bufpos] == '\t' ||
	       buf[bufpos] == '\n') {
		break;
	    }
	    else if(buf[bufpos] == '\\') {
		++bufpos;
	    }
	    name[namepos++]=buf[bufpos++];
	}
	name[namepos]=0;
	eat_ws(&bufpos, buf, 1);
	options=&buf[bufpos];
	while(bufpos < 1024 && buf[bufpos] != '\n')
	    ++bufpos;
	buf[bufpos]=0;
	L_append(l, name, options);
    }
    fclose(fp);

    if(l->list_name)
	free(l->list_name);
    l->list_name=strdup(filename);
    return 1;
}


int save_playlist(char *filename, struct list_head *l)
{
    FILE *fp;

    if(!(fp=fopen(filename, "w")))
	return 0;
    fprintf(fp, MOD_MAGIC);
    fprintf(fp, "#\n# Not intended for manual editing. "
	    "Any inconsistency may crash mod.\n\n");
    for(L_iterate_reset(l); L_iterate_valid(l); L_iterate_forward(l)) {
	fprintf(fp, "%s %s\n", escape_name(l->current_iter->filename),
		l->current_iter->options);
    }
    fclose(fp);

    if(l->list_name)
	free(l->list_name);
    l->list_name=strdup(filename);
    return 1;
}

/* Functions below are not to be called by the user */

struct list_item *L_create_item(char *filename, char *options)
{
    struct list_item *i;
    
    i=(struct list_item *)malloc(sizeof(struct list_item));
    i->filename=strdup(filename);
    i->options=strdup(options);
    return i;
}

void L_insert_first(struct list_head *l, struct list_item *i)
{
    i->next=i->prev=0;
    if(!l->first) {
	l->first=l->last=i;
    }
    else {
	i->next=l->first;
	l->first->prev=i;
	l->first=i;
    }
    ++l->size;
}


void L_insert_last(struct list_head *l, struct list_item *i)
{
    i->next=i->prev=0;
    if(!l->first) {
	l->first=l->last=i;
    }
    else {
	i->prev=l->last;
	l->last->next=i;
	l->last=i;
    }
    ++l->size;
}


void L_insert_middle(struct list_head *l, int nr, struct list_item *new_i)
{
    struct list_item *i;

    for(i=l->first; nr; i=i->next, --nr)
	;
    new_i->prev=i;
    new_i->next=i->next;
    i->next=new_i;
    new_i->next->prev=new_i;
    ++l->size;
}
