/*
 *  load.c - Handles uncompressing and calls loaders for different formats.
 *
 *  (C) 1994, 1995 Mikael Nordqvist (d91mn@efd.lth.se, mech@df.lth.se)
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <sys/ultrasound.h>
#include <limits.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

#include "mod.h"

/* External global variables */

SEQ_DECLAREBUF();

/* Constants */

#define COMPRESSED_GZIP 1
#define COMPRESSED_LHA  2
#define COMPRESSED_ZIP  3
#define COMPRESSED_ARJ  4

/* Loads a module and returns nonzero if M is valid after loading. */

int load_module(char *filename)
{
    char orig_dir[PATH_MAX+1];
    int tmp;
    
    zero_resources();

    getcwd(orig_dir, PATH_MAX);
    tmp=do_load_module(filename);
    chdir(orig_dir);
    return tmp;
}


int do_load_module(char *filename)
{
    int fd, fdi, i, tmp, suffix=0, filetype=0, pos, infolen=0;
    char buf[PATH_MAX+1]; /* Should actually be a little longer but... */
    char infofilename[PATH_MAX+1];
    struct stat sa;

    infofilename[0]=0;
    
    /* Set M.filename to modname w/o path */
    strcpy(M.real_filename, filename);
    tmp=strlen(M.real_filename);
    for(i=tmp-1; i >= 0 && M.real_filename[i] != '/'; )
	--i;
    strcpy(M.filename, &M.real_filename[++i]);
    tmp=strlen(M.filename);
    
    if(tmp > 2) {
	if(!strcmp(M.filename+tmp-2, ".z") ||
	   !strcmp(M.filename+tmp-2, ".Z")) {
	    filetype=COMPRESSED_GZIP;
	    suffix=2;
	}
    }
    if(tmp > 3) {
	if(!strcmp(M.filename+tmp-3, ".gz")) {
	    filetype=COMPRESSED_GZIP;
	    suffix=3;
	}
    }
    if(tmp > 4) {
	if(!strcmp(M.filename+tmp-4, ".lha") ||
	   !strcmp(M.filename+tmp-4, ".lzh")) {
	    filetype=COMPRESSED_LHA;
	    suffix=4;
	}
	else if(!strcmp(M.filename+tmp-4, ".zip")) {
	    filetype=COMPRESSED_ZIP;
	    suffix=4;
	}
	else if(!strcmp(M.filename+tmp-4, ".arj")) {
	    filetype=COMPRESSED_ARJ;
	    suffix=4;
	}
    }
    
    M.filename[tmp-suffix]=0; /* Remove suffix */
    
    /* Make sure the file exists by opening and closing the file */
    if(!(fd=open(M.real_filename, O_RDONLY, 0))) {
	print_status("Unable to read file");
	info("Unable to open file '%s'.\n", M.real_filename);
	sleep(1);
	return 0;
    }
    close(fd);
    
    if(!stat(M.real_filename, &sa))
	M.filesize=sa.st_size;
    else
	M.filesize=0; /* Failed stat ? Shouldn't happen... */
    
    /* Print status if we are uncompressing */
    switch(filetype) {
      case COMPRESSED_GZIP:
      case COMPRESSED_LHA:
      case COMPRESSED_ZIP:
      case COMPRESSED_ARJ:
	print_status("Uncompressing module...");
	info("Uncompressing module... ");
	break;
      default:
    }

    /* Prepare uncompress command */
    switch(filetype) {
      case COMPRESSED_GZIP:
	pos=sprintf(buf, GZIP_COMMAND " %s > " TMP_DIR "/",
		    escape_name(M.real_filename));
	sprintf(&buf[pos], "%s 2>/dev/null < /dev/null",
		escape_name(M.filename));
	break;
      case COMPRESSED_LHA:
	infolen=guess_lha_filename(infofilename);
	chdir(TMP_DIR);
	pos=sprintf(buf, LHA_COMMAND " %s", escape_name(M.real_filename));
	pos+=sprintf(&buf[pos], " %s", escape_name(M.filename));
	if(infofilename[0])
	    sprintf(&buf[pos], " %s", escape_name(infofilename));
	strcat(buf, " >& /dev/null < /dev/null");
	break;
      case COMPRESSED_ZIP:
	infolen=guess_zip_filename(infofilename);
	chdir(TMP_DIR);
	pos=sprintf(buf, ZIP_COMMAND " %s", escape_name(M.real_filename));
	pos+=sprintf(&buf[pos], " %s", escape_name(M.filename));
	if(infofilename[0])
	    sprintf(&buf[pos]," %s", escape_name(infofilename));
	strcat(buf, " >& /dev/null < /dev/null");
	break;
      case COMPRESSED_ARJ:
	infolen=guess_arj_filename(infofilename);
	chdir(TMP_DIR);
	sprintf(buf, ARJ_COMMAND " %s >& /dev/null < /dev/null",
		escape_name(M.real_filename));
	break;
      default:
    }

    /* Uncompress and open file (+ read infofile if there is one) */
    switch(filetype) {
      case COMPRESSED_GZIP:
      case COMPRESSED_LHA:
      case COMPRESSED_ZIP:
      case COMPRESSED_ARJ:
	tmp=my_system(buf);
	sprintf(buf, TMP_DIR "/%s", M.filename);
	if(tmp) {
	    unlink(buf);
	    if(tmp > 0) {
		print_status("Error uncompressing");
		sleep(1);
	    }
	    return 0;
	}
	
	/* Open uncompressed module */
	if((fd=open(buf, O_RDONLY, 0)) == -1) {
	    info("Unable to open file '%s'.\n", buf);
	    return 0;
	}

	/* Read infofile if there is one */
	if(infofilename[0]) {
	    if((fdi=open(infofilename, O_RDONLY, 0)) != -1) {
		M.infofile=(char *)malloc(infolen+1);
		if((tmp=read(fdi, M.infofile, infolen)) > 0) {
		    M.infofile[tmp]=0;
		}
		else {
		    free(M.infofile);
		    M.infofile=0;
		}
		close(fdi);
		unlink(infofilename);
	    }
	}
	break;
      default: /* Assume non-compressed module */
	if((fd=open(M.real_filename, O_RDONLY, 0)) == -1) {
	    print_status("File does not exist");
	    info("Unable to open file '%s'.\n", M.real_filename);
	    sleep(1);
	    return 0;
	}
    }

    /* Here and forward buf should contain the name of the temp-file */
    chmod(buf, 0777); /* a+rwx */
    
    /* Try to determine module-format from the filename if needed */
    if(!opt.format) {
	if(strlen(M.filename) > 4) {
	    if(is_modtype("mod") || is_modtype("nst"))
		opt.format=MODFORMAT_MOD;
	    else if(is_modtype("ult"))
		opt.format=MODFORMAT_ULT;
	    else if(is_modtype("mtm"))
		opt.format=MODFORMAT_MTM;
	    else if(is_modtype("s3m"))
		opt.format=MODFORMAT_S3M;
	}
	if(!opt.format)
	    opt.format=MODFORMAT_MOD; /* Default to MOD-format */
    }
    
    M.format=opt.format;
    M.sample=0;
    
    switch(opt.format) {
      case MODFORMAT_MOD:
	/* Printing done in load_mod() as there are several MOD-types */
	tmp=load_mod(fd); /* MOD */
	break;
      case MODFORMAT_ULT:
	print_status("Loading ULT...");
	tmp=load_ult(fd); /* ULT */
	break;
      case MODFORMAT_MTM:
	print_status("Loading MTM...");
	tmp=load_mtm(fd); /* MTM */
	break;
      case MODFORMAT_S3M:
	print_status("Loading S3M...");
	tmp=load_s3m(fd); /* S3M */
	break;
      default:
	error("Error selecting format.\n");
    }
    close(fd);

    info("\n\n");
    
    if(M.nr_samples && M.sample)
	M.sample[0].valid=0;  /* Make sure sample 0 is invalid */

    /* Remove temp-files if module was compressed */
    if(filetype) {
	unlink(buf);
    }

#if 0
    info("\nTracks stored: %d/%d.\n", M.nr_tracks, M.nr_voices*M.nr_patterns);
#endif

    if(tmp) {
	if(opt.start_pos >= M.songlength)
	    opt.start_pos=0;
    }

    return tmp;      /* Return status from load_xxx */
}


int is_modtype(char *type)
{
    char buf[5], name[256];
    int i;
    
    strcpy(name, M.filename);
    for(i=0; name[i]; ++i)
	if(isalpha(name[i]) && isupper(name[i]))
	    name[i]=tolower(name[i]);

    i=strlen(name)-4;
    
    buf[0]='.';
    strcpy(buf+1, type);
    if(!strncmp(name+i, buf, 4))        /* .xxx */
	return 1;
    buf[0]='_';
    if(!strncmp(name+i, buf, 4))        /* _xxx */
	return 1;

    strcpy(buf, type);
    strcat(buf, ".");
    if(!strncmp(name, buf, 4))          /* xxx. */
	return 1;
    buf[3]='_';
    if(!strncmp(name, buf, 4))          /* xxx_ */
	return 1;
    
    return 0;
}


int guess_lha_filename(char *infofilename)
{
    char buf[PATH_MAX+1], maxname[PATH_MAX+1], minname[PATH_MAX+1];
    int maxsize, minsize, namep, sizep, nr_files;
    FILE *fp;
    
    sprintf(buf, LHA_COMMAND_LIST " %s 2> /dev/null < /dev/null",
	    escape_name(M.real_filename));
    if(!(fp=popen(buf, "rb")))
	return 0;

    strcpy(maxname, M.filename); /* Default to name minus suffix */
    minname[0]=0;
    
    maxsize=0;
    minsize=10240; /* File 'has' to be smaller than 10k to be a textfile*/
    nr_files=0;
    
    /* Find the largest/smallest file */
    if(fgets(buf, PATH_MAX, fp) && fgets(buf, PATH_MAX, fp)) {
	while(!feof(fp)) {
	    if(!fgets(buf, PATH_MAX, fp))
		break;
	    if(strlen(buf) <= 1 || *buf == '-')
		break;
	    
	    buf[strlen(buf)-1]=0; /* Remove lf */

	    /* Start from %-sign in output */
	    for(namep=0; buf[namep] && buf[namep] != '%'; ++namep)
		;
	    if(!buf[namep]) /* Abort if no '%' found */
		break;
	    
	    sizep=namep++;

	    /* Locate filename */
	    eat_ws(&namep, buf, 1);
	    eat_nonws(&namep, buf, 1);
	    eat_ws(&namep, buf, 1);
	    eat_nonws(&namep, buf, 1);
	    eat_ws(&namep, buf, 1);
	    eat_nonws(&namep, buf, 1);
	    eat_ws(&namep, buf, 1);
	    
	    /* Locate filesize */
	    eat_nonws(&sizep, buf, -1);
	    eat_ws(&sizep, buf, -1);
	    eat_nonws(&sizep, buf, -1);
	    sizep++;

	    ++nr_files;
	    
	    if(atoi(&buf[sizep]) > maxsize) {
		maxsize=atoi(&buf[sizep]);
		strcpy(maxname, &buf[namep]);
	    }
	    
	    if(atoi(&buf[sizep]) < minsize) {
		minsize=atoi(&buf[sizep]);
		strcpy(minname, &buf[namep]);
	    }
	}
    }
    pclose(fp);
    strcpy(M.filename, maxname);
    if(nr_files >= 2 && strcmp(maxname, minname) && minname[0])
	strcpy(infofilename, minname);
    return minsize;
}


int guess_zip_filename(char *infofilename)
{
    char buf[PATH_MAX+1], maxname[PATH_MAX+1], minname[PATH_MAX+1];
    int maxsize, minsize, nr_files;
    FILE *fp;
    
    sprintf(buf, ZIP_COMMAND_LIST " %s 2> /dev/null < /dev/null",
	    escape_name(M.real_filename));
    if(!(fp=popen(buf, "rb")))
	return 0;
    
    strcpy(maxname, M.filename); /* Default to name minus suffix */
    minname[0]=0;

    maxsize=0;
    minsize=10240; /* File 'has' to be smaller than 10k to be a textfile*/
    nr_files=0;
    
    /* Find the largest file */
    if(fgets(buf, PATH_MAX, fp) && fgets(buf, PATH_MAX, fp)) {
	while(!feof(fp)) {
	    if(!fgets(buf, PATH_MAX, fp))
		break;
	    if(strlen(buf) <= 3 || !strncmp(buf, " -", 2))
		break;
	    
	    buf[strlen(buf)-1]=0; /* Remove lf */
	    
	    ++nr_files;
	    
	    if(atoi(buf) > maxsize) {
		maxsize=atoi(buf);
		strcpy(maxname, &buf[27]);
	    }
	    if(atoi(buf) < minsize) {
		minsize=atoi(buf);
		strcpy(minname, &buf[27]);
	    }
	}
    }
    pclose(fp);
    strcpy(M.filename, maxname);
    if(nr_files >= 2 && strcmp(maxname, minname) && minname[0])
	strcpy(infofilename, minname);
    return minsize;
}


int guess_arj_filename(char *infofilename)
{
    char buf[PATH_MAX+1], maxname[PATH_MAX+1], minname[PATH_MAX+1], buf2[20];
    int maxsize, minsize, nr_files;
    int i, size=0; /* Remove warning */
    char *tmp;
    FILE *fp;
    
    sprintf(buf, ARJ_COMMAND_LIST " %s 2> /dev/null < /dev/null",
	    escape_name(M.real_filename));
    if(!(fp=popen(buf, "rb")))
	return 0;
    
    strcpy(maxname, M.filename); /* Default to name minus suffix */
    minname[0]=0;
    
    maxsize=0;
    minsize=10240; /* File 'has' to be smaller than 10k to be a textfile*/
    nr_files=0;

    /* Find the largest file */
    tmp=(char *)42;
    for(i=0; tmp && i < 6; ++i) {
	tmp=fgets(buf, PATH_MAX, fp);
	if(!tmp)
	    break;
    }
    if(tmp) {
	while(!feof(fp)) {
	    if(!fgets(buf, PATH_MAX, fp))
		break;
	    if(strlen(buf) <= 14 || !strncmp(buf, "------------ -", 14))
		break;

	    buf[strlen(buf)-1]=0; /* Remove lf */
	    size=atoi(&buf[13]);
	    
	    /* Fix filename, dos-way (8.3) + lowercase for unix unarj */
	    strncpy(buf2, buf, 12);
	    for(i=11; i >= 0 && buf2[i] == ' '; --i)
		;
	    buf2[++i]=0;
	    for(i=0; buf2[i]; ++i)
		if(isalpha(buf2[i]) && isupper(buf2[i]))
		    buf2[i]=tolower(buf2[i]);

	    ++nr_files;

	    if(size > maxsize) {
		maxsize=size;
		strcpy(maxname, buf2);
	    }
	    if(size < minsize) {
		minsize=size;
		strcpy(minname, buf2);
	    }
	}
    }
    pclose(fp);
    strcpy(M.filename, maxname);
    if(nr_files >= 2 && strcmp(maxname, minname) && minname[0])
	strcpy(infofilename, minname);
    return size;
}


void print_used_effects(void)
{
    int v, l;

    if(opt.verbose < 2) {
	info("\n");
	return;
    }

    info("\nEFX: ");
    l=0;
    for(v=0; v < NR_EFX; ++v)
	if(effect_used[v]) {
	    info("%s ",effectnames[v]);
	    ++l;
	}
    info("\n\n");
#if 0
    info("(# of EFX: %d)\n",l);
#endif
}


void free_module(void)
{
    int i;
    
    for(i=0; i < MAX_VOICES; ++i)
	if(M.track_idx[i])
	    free(M.track_idx[i]);
    
    if(M.tracks) {
	for(i=0; i < M.nr_tracks; ++i)
	    free(M.tracks[i]);
	free(M.tracks);
    }
    
    if(M.songtext)
	free(M.songtext);
    
    if(M.infofile)
	free(M.infofile);
    
    if(M.sample)
	free(M.sample);
    
    zero_resources();
}


void zero_resources(void)
{
    int v;
    
    for(v=0; v < MAX_VOICES; ++v)
	M.track_idx[v]=0;
    
    M.tracks=0;
    M.sample=0;
    M.songtext=0;
    M.infofile=0;
}
