/*
 *  main.c - Initialization before main-loop and cleanup before exit.
 *           Handling of signals.
 *
 *  (C) 1994, 1995 Mikael Nordqvist (d91mn@efd.lth.se, mech@df.lth.se)
 */

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <errno.h>
#include <sys/soundcard.h>
#include <sys/stat.h>
#include <signal.h>
#include <string.h>
#include <time.h>

#include "mod.h"

SEQ_DECLAREBUF();

int main(int argc, char *argv[])
{
    exe_name=argv[0];
    GUI_mode=GUI_SIMPLE;
    quiet=loaded=quit=suspend=0;
    files=0;
    file_names=0;
    mixer_name=DEFAULT_MIXER;
    playlist_name[0]=0;
    
    srand(time((time_t)0));    /* Randomize */
    
    if(mkdir(TMP_DIR , 0777) && errno != EEXIST)
	error("Unable to create " TMP_DIR ".\n");
    chmod(TMP_DIR, 0777); /* This is needed as mode&umask is used in mkdir */
    
    if(argc >= 2 && !strncmp(argv[1], "-k", 2))
	kill_mod();
#if 0    
    /* Special case to allow help even if mod is already running */
    if(argc >= 2 && !strncmp(argv[1], "-h", 2)) {
	print_helptext();
	exit(0);
    }
#endif

    getcwd(workdir, PATH_MAX); /* Initialize working directory */

#ifdef COMPILE_GUI_TK
    /* Upgrade to GUI_TK if possible */
    GUI_size=3; /* Default to largest window w/o samples */
    if(tk_ok()) {
	GUI_mode=GUI_TK;
	tk_parse_options(&argc, argv); /* Parse TK-options */
    }
#endif
#ifdef COMPILE_GUI_NC
    /* Upgrade to GUI_NC if possible (tk_ok() failed) */
    if(GUI_mode < GUI_NC && nc_ok())
	GUI_mode=GUI_NC;
#endif
    init_options(argc, argv);

    pid_main=getpid();
    maxfd=getdtablesize();
    
    /* Create communication-pipes */
    if(pipe(player2main_synced) || pipe(main2player) || pipe(ack_pipe))
        error("Unable to create pipes.");
    
    /* Player should never ever block. Neither should the keyboard. */
    fcntl(player2main_synced[PIPE_WRITE], F_SETFL, O_NONBLOCK);
    fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);
    
    init_sound();  /* Make sure audio-resources are available and grab them */
    init_mixer();  /* Open mixer */
    
    if(!L_size(PL) && GUI_mode <= GUI_SIMPLE)
	error("No songs specified.\n");
    
    init_screen(); /* This one might change signal-handlers */
    
    /* Install signal handlers _after_ init_screen() */
    pid_player=1;       /* Make sure handlers know we are the main process */
    install_signalhandlers();

    draw_screen(); /* Draw the screen */

    zero_resources();  /* Make sure nothing undefined get free()'d */
    /* cleanup_sound();  Free resources before entering main loop */
    init_controller(); /* Make sure controller is ready */
    main_loop();       /* Do it! */
    
    return 0; /* Never reached */
}


void cleanup(void)
{
    cleanup_screen();

    chdir(TMP_DIR);
    chdir("..");
    my_removedir(TMP_DIR);  /* Try to delete all garbage-files in TMP_DIR */

    print_playingtime();
}


void write_pid(pid_t p)
{
    int fd;
    char buf[10];

    if((fd=open(TMP_DIR"/mod.pid", O_RDONLY)) != -1) {
	error("There is a copy of mod running in the background, try "
	      "mod -k.\n");
    }

    if(!p) /* Just checking for stale lock */
	return;

    if((fd=open(TMP_DIR"/mod.pid", O_CREAT|O_TRUNC|O_WRONLY, 0777)) == -1)
	error("Unable to write 'mod.pid'", errno);
    sprintf(buf, "%09d\n", p);
    write(fd, buf, 10);
    close(fd);
    chmod(TMP_DIR"/mod.pid", 0777);
}


void kill_mod(void)
{
    int fd, p;
    char buf[10];

    if((fd=open(TMP_DIR"/mod.pid", O_RDONLY)) == -1)
	error("No copy of mod running in the background.\n");

    if(read(fd, buf, 12) != 10)
	error("Trouble reading 'mod.pid'.\n");
    close(fd);
    p=atoi(buf);
    if(kill(p, SIGTERM)) {
	if(errno == EPERM)
	    error("You don't have permission to kill the running copy of"
		  " 'mod'.\n");
	else if(errno == ESRCH) {
	    if(!unlink(TMP_DIR"/mod.pid")) {
		printf("Stale lock removed.\n");
		exit(0);
	    }
	    else
		error("Unable to remove stale lock.");
	}
	else
	    error("No copy of mod running in the background.\n");
    }
    printf("Killed.\n");
    exit(0);
}

/* Main-process:  Set quit-flag, kill child.
 * Child-process: Terminate. Maybe we should kill() father also?
 */

void handler_quit(int sig)
{
    if(pid_player) {
	install_ignore_signalhandlers();
	if(pid_player != 1) /* Before forking of the first child */
	    kill(pid_player, sig);
#ifdef COMPILE_GUI_TK
	if(GUI_mode == GUI_TK)
	    tk_quit_signal();
	else
#endif
	    quit=QUIT_SIGNAL;
    }
    else {
	cleanup_sound();
	exit(0);
    }
}


void install_signalhandlers(void)
{
    struct sigaction sa, osa;
    
    sa.sa_handler=handler_quit;
    sa.sa_flags=0;
    sigemptyset(&sa.sa_mask);
    
    sigaction(SIGHUP, &sa, &osa);
    sigaction(SIGINT, &sa, &osa);
    sigaction(SIGQUIT, &sa, &osa);
    sigaction(SIGTERM, &sa, &osa);
#if 0    
    install_suspend_handler();
#else
    install_ignore_suspend_handler();
#endif
}


void install_ignore_signalhandlers(void)
{ 
    struct sigaction sa, osa;
   
    sa.sa_handler=SIG_IGN;
    sa.sa_flags=0;
    sigemptyset(&sa.sa_mask);
    
    sigaction(SIGHUP, &sa, &osa);
    sigaction(SIGINT, &sa, &osa);
    sigaction(SIGQUIT, &sa, &osa);
    sigaction(SIGTERM, &sa, &osa);
    sigaction(SIGTSTP, &sa, &osa);
}


/* Experimental suspend. Doesn't work yet. */

#if 0

static struct sigaction old_suspend_sa;

void suspend_program(void)
{
    sigset_t mask, omask;
    
    if(old_suspend_sa.sa_handler == SIG_IGN)
	debug("IGNORE!!!\n\n\nIGNORE\n");
    /* Reinstall old handler */
    sigaction(SIGTSTP, &old_suspend_sa, (struct sigaction *)0);
    
    /* Unblock SIGTSTP just to be sure */
    sigemptyset(&mask);
    sigaddset(&mask, SIGTSTP);
    sigprocmask(SIG_UNBLOCK, &mask, &omask);
    
    /* Ok, suspend us. */
    kill(getpid(), SIGTSTP);

    /* Done. Reinstall our own handler */
    install_suspend_handler();

    /* Restore signal mask */
    sigprocmask(SIG_SETMASK, &omask, (sigset_t *)0);
    suspend=0;
    fcntl(STDIN_FILENO, F_SETFL, O_NONBLOCK);
}


void install_suspend_handler(void)
{
    struct sigaction sa;
    
    sa.sa_handler=handler_suspend;
    sa.sa_flags=0;
    sigemptyset(&sa.sa_mask);
    sigaction(SIGTSTP, &sa, &old_suspend_sa);
}


void handler_suspend(int sig)
{
    suspend=1;
    debug("\nGot SIGTSTP\n");
    debug("\nGot SIGTSTP\n");
    debug("\nGot SIGTSTP\n");
}
#endif


void install_ignore_suspend_handler(void)
{
    struct sigaction sa;
    
    sa.sa_handler=SIG_IGN;
    sa.sa_flags=0;
    sigemptyset(&sa.sa_mask);
    sigaction(SIGTSTP, &sa, (struct sigaction *)0);
}
