/*
 *  nc_screen.c - Handling of the screen using the ncurses package.
 *
 *  (C) 1994, 1995 Mikael Nordqvist (d91mn@efd.lth.se, mech@df.lth.se)
 */

#ifdef COMPILE_GUI_NC

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <ncurses.h>

#include "mod.h"

/* Defines that attempts to circumvent color-bugs in older ncurses */

#ifdef NC_XXTRA_REFRESH
#define XXnc_wrefresh(w) nc_wrefresh(w)
#define NC_XTRA_REFRESH
#else
#define XXrefresh() ;
#define XXnc_wrefresh(w) ;
#endif

#ifdef NC_XTRA_REFRESH
#define Xrefresh() refresh
#define Xnc_wrefresh(w) nc_wrefresh(w)
#else
#define Xrefresh() ;
#define Xnc_wrefresh(w) ;
#endif

#define ATTR_W_ON_B     (COLOR_PAIR(7))

#define ATTR_LINE       (ATTR_W_ON_B)
#define ATTR_BOLDLINE   (COLOR_PAIR(7) | A_BOLD)

#define ATTR_FILE       (ATTR_W_ON_B)
#define ATTR_BOLDFILE   (ATTR_FILE | A_BOLD)
#define ATTR_DIR        (COLOR_PAIR(2) | A_BOLD)
#define ATTR_BOLDDIR    (COLOR_PAIR(4) | A_BOLD)

#define ATTR_REDBAR     (COLOR_PAIR(1) | A_BOLD)
#define ATTR_DIMREDBAR  (COLOR_PAIR(11) | A_DIM)
#define ATTR_BOLDREDBAR (COLOR_PAIR(11) | A_BOLD)

#define ATTR_VOICEON    ATTR_REDBAR
#define ATTR_VOICEOFF   ATTR_DIMREDBAR
#define ATTR_VOICEONA   ATTR_BOLDREDBAR
#define ATTR_VOICEOFFA  (COLOR_PAIR(12))

#define ATTR_PANEL     (COLOR_PAIR(3) | A_BOLD)
#define ATTR_PANELTEXT (COLOR_PAIR(9) | A_BOLD)

#define ATTR_PANEL2    (COLOR_PAIR(8) | A_BOLD)

#define ATTR_PANELBOX  (COLOR_PAIR(5) | A_BOLD)
#define ATTR_PANELBOXDIM (COLOR_PAIR(10))
#define ATTR_PANELBOX2 (COLOR_PAIR(6) | A_BOLD)

WINDOW *iwin, *swin;

int swin_LINES;

int first_voice, cur_voice, first_sample;
int nr_visible_voices;

int refresh_ok;

/* Wrapper to allow refreshes to be turned off (temporarily) */

inline void nc_wrefresh(WINDOW *w)
{
    if(refresh_ok)
	wrefresh(w);
}

/* These two functions are only used when debugging ncurses */

inline void wsetcol(WINDOW *w, chtype col)
{ 
	wattrset(w, col);
}

inline void setcol(chtype col)
{
	attrset(col);
}


int nc_ok(void)
{
#ifdef NC_CHECK_FOR_COLOR
    int res;

    initscr();
    res=has_colors();
    endwin();
    if(!res)
	return 0;
#endif
    return 1;
}

void nc_setup_color_mode(void)
{
    start_color();
    init_pair(1, COLOR_YELLOW, COLOR_RED);
    init_pair(2, COLOR_BLUE, COLOR_BLACK);
    init_pair(3, COLOR_WHITE, COLOR_CYAN);
    init_pair(4, COLOR_CYAN, COLOR_BLACK);
    init_pair(5, COLOR_YELLOW, COLOR_BLUE);
    init_pair(6, COLOR_YELLOW, COLOR_MAGENTA);
    init_pair(7, COLOR_WHITE, COLOR_BLACK);
    init_pair(8, COLOR_WHITE, COLOR_WHITE);
    init_pair(9, COLOR_BLUE, COLOR_CYAN);
    init_pair(10, COLOR_WHITE, COLOR_BLUE);
    init_pair(11, COLOR_WHITE, COLOR_RED);
    init_pair(12, COLOR_GREEN, COLOR_RED);
}

void nc_init_screen(void)
{
    initscr();

    refresh_ok=1;

#ifdef NC_CHECK_FOR_COLOR
    if(!has_colors())
	error("No colour support on this terminal\n");
#endif

    if(LINES < 19 || COLS < 80)
	error("You have to have at least a 19x80 screen "
	      "to use the ncurses interface.\n");
    
    nc_setup_color_mode();

    swin_LINES=LINES-16-1;
    
    iwin=newwin(17, COLS, 0, 0);
    scrollok(iwin, FALSE);
    leaveok(iwin, TRUE);
    
    swin=newwin(swin_LINES, COLS, 17, 0);
    scrollok(swin, FALSE);
    leaveok(swin, TRUE);
    
    swin_LINES=MIN(default_opt.max_lines, swin_LINES-1);

    meta(stdscr, TRUE);

    cbreak();
    noecho();
    keypad(stdscr, TRUE);
    curs_set(0);         /* Make cursor invisible */
}


void nc_cleanup_screen(void)
{
    delwin(iwin);
    delwin(swin);
    keypad(stdscr, FALSE);
    setcol(ATTR_W_ON_B);
    clear();
    refresh();
    curs_set(1); /* Make cursor visible again */
    endwin();
}


void nc_draw_screen(void)
{
    set_voice_detail(1); /* Start with highest detail */
    first_voice=cur_voice=0;
    
    nc_refresh_screen(0);
}


void nc_refresh_screen(int slow)
{
    int i;

    refresh_ok=0;

    sync_volume();
    
#if 0
    setcol(ATTR_LINE);
    clear();
    refresh();
    wsetcol(swin, ATTR_LINE);
    wclear(swin);
    wrefresh(swin);
#endif
    if(slow) {
	wclear(iwin);
	wclear(swin);
    }

    wsetcol(iwin, ATTR_PANEL);
    for(i=0; i < 16; ++i)
	mvwaddstr(iwin, i, 0, "                                              "
		 "                                  ");

    mvwaddstr(iwin, 1, 1, "Status                            "
	     "Type         Songpos    /     Speed");
    mvwaddstr(iwin, 2, 1, "                                  "
	     "Voices       Pattern    /     Tempo");
    mvwaddstr(iwin, 3, 1, "Song                              "
	     "Samples      Line       /");
    mvwaddstr(iwin, 4, 1, "                                  "
	     "Size");
    mvwaddstr(iwin, 5, 1, "File                              "
	     "File         Time        :");

    mvwaddstr(iwin, 7, 2, "# Samplename                    Size  Volm Loop "
	      "Tune Bits Type");

    if(loaded) {
	nc_print_samples(first_sample);

	nc_print_pos(last_pos, last_line);
	nc_print_speed(last_speed, last_tempo);
	print_songinfo(0);
	print_time(-2);
	
	nc_print_songname(M.name);
	nc_print_filename(path_basename(M.filename, 0));
    }
    else {
	nc_print_samples(-1);

	nc_print_pos(-1,0);
	nc_print_speed(-1,0);
	print_songinfo(-1);
	print_time(-1);
	
	nc_print_songname(" ");
	nc_print_filename("");
    }

    nc_print_minihelp(0);

    if(!selecting)
	nc_print_channelnumbers();
    else
	nc_print_selectingbar();

    nc_print_lowerbar();

    set_playing(playing);

    if(selecting) {
	nc_print_files(cur_filenr);
	nc_print_list(cur_listnr);
    }
    else if(playing && !default_opt.noscroll) {
	nc_print_line(last_pos, last_line);
    }
    refresh_ok=1;
    nc_wrefresh(iwin);
    nc_wrefresh(swin);
}


void nc_touch_windows(void)
{
    touchwin(iwin);
    touchwin(swin);
}

void nc_clear_screen(void)
{
    int i;
    
    wsetcol(swin, ATTR_W_ON_B);
    for(i=0; i < swin_LINES; ++i) {
	wmove(swin, i, 0);
	wclrtoeol(swin);
    }
    nc_wrefresh(swin);
}


void nc_clear_left_screen(void)
{
    int i;
    
    wsetcol(swin, ATTR_W_ON_B);
    for(i=0; i < swin_LINES; ++i) {
	wmove(swin, i, 0);
	waddstr(swin, "                                        ");
    }
    nc_wrefresh(swin);
}


void nc_inc_swin(void)
{
    if(swin_LINES < LINES-16-2) {
	wsetcol(swin, ATTR_FILE);
        if(selecting)
            mvwaddstr(swin, swin_LINES/2-(!(swin_LINES%2) ? 1 : 0), 38, "  ");
	wmove(swin, swin_LINES, 0);
	wclrtoeol(swin);
	++swin_LINES;
	nc_resize();
    }
}


void nc_dec_swin(void)
{
    if(swin_LINES > 1) {
	wsetcol(swin, ATTR_FILE);
        if(selecting)
            mvwaddstr(swin, swin_LINES/2-(!(swin_LINES%2) ? 1 : 0), 38, "  ");
	wmove(swin, swin_LINES, 0);
	wclrtoeol(swin);
	--swin_LINES;
	nc_resize();
    }
}


void nc_resize(void)
{
    if(selecting) {
	nc_print_files(cur_filenr);
	nc_print_list(cur_listnr);
    }
    else {
	if(default_opt.noscroll && playing) {
	    nc_print_line(last_pos, last_line);
	}
    }
    nc_print_lowerbar();
}


void nc_print_lowerbar(void)
{
    char buf[81], s1[40], s2[40];
    int len, i, nr;
    
    nc_print_volume();

    len=strlen(MOD_VERSION);
    nr=(80-38-len)/2;
    for(i=0; i+14 < nr; ++i)
	s1[i]=' ';
    s1[i]=0;
    nr=80-nr-38-len;
    for(i=0; i < nr-13; ++i)
	s2[i]=' ';
    s2[nr]=0;
    
    sprintf(buf, "%sMod " MOD_VERSION " - "
	    "(C) 1994, 1995 Mikael Nordqvist%s", s1, s2);
    
    wsetcol(swin, ATTR_REDBAR);
    mvwaddstr(swin, swin_LINES, 14, buf);
    nc_wrefresh(swin);

    nc_print_flags();
}


static void nc_set_channel_attr(int voice)
{
    if(opt.active_voices & (1<<voice)) {
	if(voice == cur_voice)
	    wsetcol(iwin, ATTR_VOICEONA);
	else
	    wsetcol(iwin, ATTR_VOICEON);
    }
    else {
	if(voice == cur_voice)
	    wsetcol(iwin, ATTR_VOICEOFFA);
	else
	    wsetcol(iwin, ATTR_VOICEOFF);
    }
}


void nc_print_channelnumbers(void)
{
    char buf[81];
    int i, x, y, delta, actualvoices;
    struct event dummy={0,0,0,0,0,0};
    
    if(!loaded) {
	wsetcol(iwin, ATTR_REDBAR);
	mvwprintw(iwin, 16, 0, fill_string("", 80));
    }
    else {
	/* Find out characters/channel (delta) */
	delta=print_event(buf, &dummy, nr_visible_voices);

	actualvoices=MIN(nr_visible_voices, M.nr_voices);
	wsetcol(iwin, ATTR_BOLDREDBAR);
	mvwprintw(iwin, 16, 0, "  %s %s", (first_voice ? "<<" : "  "),
		  fill_string("", (80-5-4-actualvoices*delta)/2+(delta-2)/2));
	Xnc_wrefresh(iwin);
	
	for(i=first_voice; i < first_voice+actualvoices-1; ++i) {
	    nc_set_channel_attr(i);
	    wprintw(iwin, "%2d%s", i+1, fill_string("", delta-2));
	    Xnc_wrefresh(iwin);
	}
	nc_set_channel_attr(i);
	
	getyx(iwin, y, x);
	wprintw(iwin, "%2d%s", i+1, fill_string("", 80-1-2-x-4));
	Xnc_wrefresh(iwin);

	wsetcol(iwin, ATTR_BOLDREDBAR);
	wprintw(iwin, "%s   ",	(i < M.nr_voices-1 ? ">>" : "  "));
    }
    nc_wrefresh(iwin);
}


void nc_print_selectingbar(void)
{
    char *s;

    if(!selecting) /* If called from nc_error_update_screen() */
	return;
    
    if(selecting == SELECTING_LOAD)
	wsetcol(iwin, ATTR_REDBAR);
    else
	wsetcol(iwin, ATTR_DIMREDBAR);
    
    s=trunc_filename(workdir, 30);
    fix_string(s);
    mvwprintw(iwin, 16, 0, "   %34s   ", fill_string(s, 34));
    Xnc_wrefresh(iwin);
    
    if(selecting == SELECTING_LOAD)
	wsetcol(iwin, ATTR_DIMREDBAR);
    else
	wsetcol(iwin, ATTR_REDBAR);
    
    waddstr(iwin, "Playlist                                ");
    nc_wrefresh(iwin);
}


void nc_print_songname(char *n)
{
    char noname[]="<no name>";

    if(n && !strlen(n))
	n=noname;
    wsetcol(iwin, ATTR_PANELBOX);
    mvwaddstr(iwin, 3, 6, fill_string(n, 27));
    nc_wrefresh(iwin);
}


void nc_print_filename(char *s)
{
    wsetcol(iwin, ATTR_PANELBOX);
    mvwaddstr(iwin, 5, 6, fill_string(s, 27));
    nc_wrefresh(iwin);
}


void nc_print_status(char *s)
{
    wsetcol(iwin, ATTR_PANELBOX);
    mvwaddstr(iwin, 1, 8, fill_string(s ? s : "", 25));
    nc_wrefresh(iwin);
}


void nc_print_files(int n)
{
    int i, f;
    int oldattr, newattr;
    char *s;

    f=n-swin_LINES/2;
    if(!(swin_LINES%2))
	f++;

    oldattr=ATTR_FILE;
    wsetcol(swin, oldattr);
    for(i=0; i < swin_LINES; ++i, ++f) {
	if(f < 0 || f >= nr_files) {
	    mvwaddstr(swin, i, 0, get_filestring(-1));
	}
	else {
	    if(f != n) {
		if(is_dir(f))
		    newattr=ATTR_DIR;
		else
		    newattr=ATTR_FILE;
	    }
	    else {
		if(is_dir(f))
		    newattr=ATTR_BOLDDIR;
		else
		    newattr=ATTR_BOLDFILE;
	    }
	    if(newattr != oldattr) {
		Xnc_wrefresh(swin);
		wsetcol(swin, newattr);
		oldattr=newattr;
	    }
	    s=get_filestring(f);
	    if(f == n && selecting == SELECTING_LOAD) {
		s[8]='-';
		s[9]='>';
		mvwaddstr(swin, i, 38, "  ");
	    }
	    mvwaddstr(swin, i, 0, s);
	}
    }
    nc_wrefresh(swin);
}


void nc_print_list(int n)
{
    int i, f;
    int oldattr, newattr, nr;

    f=n-swin_LINES/2;
    if(!(swin_LINES%2))
	f++;

    nr=L_size(PL);
    L_iterate_reset(PL);
    if(f > 0) {
	for(i=0; i < f; ++i)
	    L_iterate_forward(PL);
    }

    oldattr=ATTR_FILE;
    wsetcol(swin, oldattr);
    for(i=0; i < swin_LINES; ++i, ++f) {
	if(f < 0 || f >= nr || !L_iterate_valid(PL)) {
	    wmove(swin, i, 40);
	    wclrtoeol(swin);
	    if(f == n && selecting == SELECTING_LIST) {
		mvwaddstr(swin, i, 38, "->");
		mvwaddstr(swin, i, 8, "  ");
	    }
	}
	else {
	    if(f != n)
		newattr=ATTR_FILE;
	    else
		newattr=ATTR_BOLDFILE;
	    
	    if(newattr != oldattr) {
		Xnc_wrefresh(swin);
		wsetcol(swin, newattr);
		oldattr=newattr;
	    }
	    if(f == n && selecting == SELECTING_LIST) {
		mvwaddstr(swin, i, 38, "->");
		mvwaddstr(swin, i, 8, "  ");
	    }
	    mvwaddstr(swin, i, 40,
		      fill_string(path_basename(PL->current_iter->filename,
					   default_opt.dir_components),40));
	    L_iterate_forward(PL);
	}
    }
    nc_wrefresh(swin);
}


void nc_print_samples(int n)
{
    int i;
    struct sample_info *s;
#ifdef FLUSH_INPUT
    char buf[80];
#endif
    
    wsetcol(iwin, ATTR_PANELBOX);
    for(i=0; i < 8; ++i) {
	if(n == -1)
	    mvwaddstr(iwin, 8+i, 1,
		      "                                                "
		      "               ");
	else {
	    s=&M.sample[n];
	    if(s->length > 4)
		wsetcol(iwin, ATTR_PANELBOX);
	    else
		wsetcol(iwin, ATTR_PANELBOXDIM);

	    mvwprintw(iwin, 8+i, 1,
		      "%2d %s %6d  %3d   %s %+2d   %2s  %-4s",
		      n, fill_string((M.format == MODFORMAT_ULT ? s->dosname :
				      s->name), 27),
		      s->length, s->volume,
		      (s->looped ?
		       (s->looped == LOOP_BIDIR ? "BIDI" : "Yes ") : "    "),
		      s->finetune, (s->bits_16 ? "16" : " 8"),
		      s->length > 4 ? "mono" : "");
	}
	Xnc_wrefresh(iwin);
	if(n != -1 && ++n > M.nr_samples)
	    n=-1;
    }
    nc_wrefresh(iwin);
#ifdef FLUSH_INPUT
    read(STDIN_FILENO, buf, 80); /* Flush input (for slow CPU's) */
#endif
}


void nc_pre_load_update_screen(void)
{
    if(!default_opt.browse || !selecting) {
	nc_print_channelnumbers();
	nc_clear_screen();
    }
    nc_clear_all_info();
}


void nc_post_load_update_screen(void)
{
    first_sample=1;
    first_voice=cur_voice=0;
    set_voice_detail(default_opt.voice_detail);
    if(!default_opt.browse || !selecting) {
	nc_clear_screen();
	nc_print_samples(first_sample);
	nc_print_channelnumbers();
    }
}


void nc_error_load_update_screen(void)
{
    nc_print_selectingbar();
}


#define NR_HELP_PAGES 10

char *help_pages[NR_HELP_PAGES][9]= {
    {
	"  Welcome!!  ",
	"             ",
	" I hope you  ",
	" will enjoy  ",
	" using mod!  ",
	"             ",
	"F1 and h will",
	"let you flip ",
	" help pages. "
    },
    {
	"   Always:   ",
	"             ",
	" h  Help     ",
	" l  Load-mode",
	"+/- Volume   ",
	" m  Muting   ",
	">/< Number of",
        "    lines    ",
	" q  Quit     "
    },
    {
	"   Flags:    ",
	"             ",
	" b  Browse   ",
	"             ",
	" r  Random   ",
	"    mode     ",
	"             ",
	" s  Scrolling",
	"             "
    },
    {
	" Load-mode:  ",
	"             ",
	"n/p Next/prev",
	"    file     ",
	"N/P Next/prev",
	"    file page",
	"TAB Toggle   ",
	"    between  ",
	"    file/list"
    },
    {
	" Load-mode:  ",
	"             ",
	"ENT Load     ",
	"    file/.lst",
	"             ",
	" S  Save     ",
	"    playlist ",
	"             ",
	"             "
    },
    {
	"File select: ",
	"             ",
	" u  Parent   ",
	"    directory",
	"t/T Tag file ",
	"    for list ",
	"DEL Remove   ",
	"    file from",
	"    list     "
    },
    {
	"File select: ",
	"             ",
	" D  Delete a ",
	"    file     ",
	"             ",
	"(requires -Z)",
	"             ",
	"             ",
	"             ",
    },
    {
	" Otherwise:  ",
	"             ",
	"PgUp FF      ",
	"PgDn RW      ",
	"HOME Restart ",
	"             ",
	"n/p Next/prev",
        "    song     ",
	"ENT Play/Stop"
    },
    {
	" Otherwise:  ",
	"             ",
	" v  Voice    ",
	"    detail   ",
	" .  Toggle   ",
	"    one voice",
	" x  Toggle   ",
        "    all      ",
	"    voices   "
    },
    {
	"  There are  ",
	"  more keys  ",
	"  available. ",
	"(up/down etc)",
	"             ",
        "Read The Fine",
	" Manual for  ",
	"    more     ",
	" information."
    }
};

void nc_print_minihelp(int change)
{
    static int cur_page=0;
    int i;

    if(change)
	cur_page=(cur_page+NR_HELP_PAGES+change)%NR_HELP_PAGES;

    wsetcol(iwin, ATTR_REDBAR);

    wmove(iwin, 5, 65);
    wvline(iwin, ACS_VLINE, 9);
    wmove(iwin, 5, 79);
    wvline(iwin, ACS_VLINE, 9);

    wmove(iwin, 4, 65);
    waddch(iwin, ACS_ULCORNER);
    whline(iwin, ACS_HLINE, 13);
    wmove(iwin, 4, 79);
    waddch(iwin, ACS_URCORNER);

    wmove(iwin, 14, 65);
    waddch(iwin, ACS_LLCORNER);
    whline(iwin, ACS_HLINE, 13);
    wmove(iwin, 14, 79);
    waddch(iwin, ACS_LRCORNER);

    for(i=0; i < 9; ++i)
	mvwaddstr(iwin, 5+i, 66,  help_pages[cur_page][i]);
    
    nc_wrefresh(iwin);
}

#ifdef DISPLAY_SEQ_IO_Q
#include <sys/ioctl.h>
#include <sys/soundcard.h>

extern int seqfd;
#endif

void nc_print_pos(int pos, int line)
{
    int i;
#ifdef DISPLAY_SEQ_IO_Q
    int o;
    char buf[80];
#endif

    wsetcol(iwin, ATTR_PANELTEXT); 
    if(pos != -1) {
	mvwprintw(iwin, 1, 56, "%3d", pos);
	mvwprintw(iwin, 2, 56, "%3d", M.patterntable[pos]);
	mvwprintw(iwin, 3, 56, "%3d", line);
    }
    else {
	for(i=1; i <= 3; ++i)
	    mvwprintw(iwin, i, 56, "   ");
    }
    nc_wrefresh(iwin);

#ifdef DISPLAY_SEQ_IO_Q
    ioctl(seqfd, SNDCTL_SEQ_GETINCOUNT, &i);
    ioctl(seqfd, SNDCTL_SEQ_GETOUTCOUNT, &o);
    sprintf(buf, "I/O: %d/%d", i, 1024-o);
    nc_print_status(buf);
#endif
}

void nc_print_speed(int spd, int tempo)
{
    wsetcol(iwin, ATTR_PANELTEXT);
    if(spd != -1) {
	mvwprintw(iwin, 1, 71, "%3d", spd);
	mvwprintw(iwin, 2, 71, "%3d", tempo);
    }
    else {
	mvwprintw(iwin, 1, 71, "   ");
	mvwprintw(iwin, 2, 71, "   ");
    }
    nc_wrefresh(iwin);
}


void nc_print_songinfo(int v, int samp, char *type, int sz, int filesz, int songlen, int pat)
{
    int i;

    wsetcol(iwin, ATTR_PANELTEXT);

    if(v != -1) {
	mvwprintw(iwin, 1, 43, "%3s", type);
	mvwprintw(iwin, 2, 44, "%2d", v);
	mvwprintw(iwin, 3, 44, "%2d", samp);
	mvwprintw(iwin, 4, 41, "%4dk", sz/1024);
	mvwprintw(iwin, 5, 41, "%4dk", filesz/1024);

	mvwprintw(iwin, 1, 60, "%3d", songlen-1);
	mvwprintw(iwin, 2, 60, "%3d", pat-1);
	mvwprintw(iwin, 3, 60, " 63");
    }
    else {
	for(i=1; i <= 3; ++i)
	    mvwprintw(iwin, i, 43, "   ");
	mvwprintw(iwin, 4, 41, "     ");
	mvwprintw(iwin, 5, 41, "     ");
	for(i=1; i <= 3; ++i)
	    mvwprintw(iwin, i, 60, "   ");
    }
    nc_wrefresh(iwin);
}    


void nc_print_time(int t)
{
    wsetcol(iwin, ATTR_PANELTEXT);
    if(t != -1) {
	mvwprintw(iwin, 5, 57, "%3d", t/60);
	mvwprintw(iwin, 5, 61, "%02d", t%60);
    }
    else {
	mvwprintw(iwin, 5, 57, "   ");
	mvwprintw(iwin, 5, 61, "  ");
    }
    nc_wrefresh(iwin);
}


void nc_clear_all_info(void)
{
    nc_print_samples(-1);
    nc_print_status("");
    nc_print_songname(" ");
    nc_print_filename("");

    nc_print_pos(-1, 0);
    nc_print_speed(-1, 0);
    nc_print_songinfo(-1, 0, 0, 0, 0, 0, 0);
    nc_print_time(-1);
}


void nc_print_line(int songpos, int line)
{
    static char tmpbuf[256]="                                        "
	"                                        ";
    char *buf;
    int pos, pat, step, v, l, i;
    char midline[81];
    int mid_i=0; /* Remove warning */

    if(default_opt.noscroll || selecting)
	return;
    
    scrollok(swin, FALSE);
    wsetcol(swin, ATTR_LINE);
    pat=M.patterntable[songpos];
    
    l=line-swin_LINES/2;
    if(!(swin_LINES%2))
	l++;
    
    buf=&tmpbuf[80];
    for(i=0; i < swin_LINES; ++i, ++l) {
	wmove(swin, i, 0);
	if(l < 0 || l > 63) {
	    wclrtoeol(swin);
	}
	else {
	    sprintf(buf, "%02d  ", l);
	    pos=4;
	    for(v=first_voice;
		v < M.nr_voices && v < first_voice+nr_visible_voices; ++v) {
		step=print_event(&buf[pos], GET_EVENT_PTR(v, pat, l),
				 nr_visible_voices);
		pos+=step;
	    }
	    sprintf(&buf[pos], "|  %02d", l);
	    pos+=5;

	    if(l != line) {
		waddstr(swin, &buf[(pos-80)/2]);
		wclrtoeol(swin);
	    }
	    else {
		strcpy(midline, &buf[(pos-80)/2]);
		mid_i=i;
	    }
	}
    }
    nc_wrefresh(swin);
    wsetcol(swin, ATTR_BOLDLINE);
    wmove(swin, mid_i, 0);
    waddstr(swin, midline);
    wclrtoeol(swin);
    nc_wrefresh(swin);
}


/* S between 1 and 3. If 0 and loaded we select a "good" one */

void set_voice_detail(int s)
{
    int detail[]={4, 9, 17};

    if(!s) {
        if(loaded) {
            if(M.nr_voices <=4)
                s=1;
            else if(M.nr_voices <=9)
                s=2;
            else
                s=3;
        }
        else
            s=1;
    }  
    nr_visible_voices=detail[s-1];
}

void nc_print_volume(void)
{
    int v;
#if 1
    int i;
    char buf[15];
#endif

    wsetcol(swin, ATTR_BOLDREDBAR);
    if(is_mute()) {
	mvwaddstr(swin, swin_LINES, 0, " [   MUTE   ] ");
    }
    else {
	v=(get_mastervolume()+9)/10;
#if 1
	sprintf(buf, " [");
	for(i=2; i-2 < v; ++i)
	    buf[i]=(i-1 == v ? '>' : '-');
	for(; i < 12; ++i)
	    buf[i]=' ';
	strcpy(&buf[i], "] ");
	mvwaddstr(swin, swin_LINES, 0, buf);
#else
	mvwaddstr(swin, swin_LINES, 0, " [          ] ");
	wmove(swin, swin_LINES, 2);
	if(v > 1)
	    whline(swin, ACS_HLINE, v-1);
	if(v > 0)
	    mvwaddch(swin, swin_LINES, 1+v, ACS_LARROW); /* Should be RARROW */
#endif
    }
    nc_wrefresh(swin);
}

void nc_print_flags(void)
{
    char buf[14]="       [   ] ";
    
    if(default_opt.browse)
	buf[8]='B';
    if(default_opt.random_mode)
	buf[9]='R';
    if(!default_opt.noscroll)
	buf[10]='S';

    wsetcol(swin, ATTR_BOLDREDBAR);
    mvwaddstr(swin, swin_LINES, 80-13, buf);
    nc_wrefresh(swin);
}

#endif /* COMPILE_GUI_NC */
