/*
 *  options.c - Handles commandline options and selection of next module.
 *
 *  (C) 1994, 1995 Mikael Nordqvist (d91mn@efd.lth.se, mech@df.lth.se)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>

#include "mod.h"

static char **av; /* Local pointer to argv */
static int ac;    /* Local copy of argc    */
static char background=0, totallyQ=0;

static char tmp_buf[2048], *tmp_av[80];  /* Should be enough */
static int tmp_ac;
static char options_str[512];

#define OPTIONS_CHECK     0     /* Only check validity */
#define OPTIONS_SET       1     /* Only set non-"always-global" options */
#define OPTIONS_SETGLOBAL 2     /* Only set "always-global" options */

void init_options(int argc, char *argv[])
{
    int fd, idx, tmp;
    char *env, *s;
    char buf[PATH_MAX+1], origdir[PATH_MAX+1];
    int pathlen;
    
    getcwd(origdir, PATH_MAX);
    
    reset_options(&default_opt); /* Compiled in defaults */
    strcpy(startup_dir, ".");
    load_rcfile();  /* Load defaults from users configfile */
    
    /* Check validity and set the "always global" options. First check
     * the environment-variable, then the commandline.
     */
    
    /* Check environment */
    if((env=getenv(ENV_VAR_NAME))) {
	if(env[0]) {
	    if(env[0] != '-')
		error("Environment-variable " ENV_VAR_NAME
		      " must begin with a hyphen.\n");
	    
	    str2array(env);
	    av=tmp_av;
	    ac=tmp_ac;
	    process_options(&default_opt, OPTIONS_SETGLOBAL, 0);
	    process_options(&default_opt, OPTIONS_SET, 0);
	}
    }
    
    /* Global section of commandline */
    av=argv;
    ac=argc;
    idx=process_options(&default_opt, OPTIONS_SETGLOBAL, 1);
    idx=process_options(&default_opt, OPTIONS_SET, 1);

    chdir(expand_name(startup_dir));

    /* Check the rest of the options and create the playlist */
    PL=&Playlist;
    L_init(PL);
    if(idx != -1) {
	/* getcwd(buf, PATH_MAX); */
	strcpy(buf, origdir);
	strcat(buf, "/");
	pathlen=strlen(buf);
	do {
	    tmp=process_options(&default_opt, OPTIONS_SETGLOBAL, idx+1);
	    if(*(av[idx]) != '/') {
		strcpy(&buf[pathlen], av[idx]);
		L_append(PL, buf, options_str);
	    }
	    else
		L_append(PL, av[idx], options_str);
	    idx=tmp;
	}
	while(idx != -1);
    }
    
    /* If the list contains just one entry that is a directory, use this
     * directory as startup directory.
     * If it ends with '.lst' it is assumed to be a playlist and loaded.
     */
    
    if(L_size(PL) == 1) {
	s=L_get_first(PL)->filename;
	if(is_dir_name(s)) {
	    strcpy(startup_dir, s);
	    chdir(expand_name(startup_dir));
	    L_clear(PL);
	}
	else if(strlen(s) >= 4 && !strcmp(".lst", &s[strlen(s)-4])) {
	    strcpy(buf, s);
	    L_clear(PL);
	    if(!load_playlist(buf, PL)) {
		printf("The list will be saved as '%s'.\n", buf);
		sleep(1);
	    }
	    strcpy(playlist_name, buf);
	}
    }
    else if(!L_size(PL)) { /* Try to load default list */
	strcpy(buf, expand_name("~/.mod.lst"));
	if(load_playlist(buf, PL))
	    strcpy(playlist_name, buf);
    }
    
    if(background && default_opt.verbose)
	error("Can't put a verbose session in the background.\n");
    
    if(quiet && default_opt.verbose)
	error("Quiet and verbose can't be used together.\n");

    if(quiet || background)
	GUI_mode=GUI_NONE;

    if((default_opt.verbose || default_opt.dump) && GUI_mode > GUI_SIMPLE)
	GUI_mode=GUI_SIMPLE;

    if(GUI_mode < GUI_NC)
	default_opt.auto_next=0;
    
    /* Make sure the selectied GUI is available */

#ifdef COMPILE_GUI_NC
    if(GUI_mode == GUI_NC && !nc_ok())
	error("No color support on this terminal.\n");
#endif
#ifdef COMPILE_GUI_TK
    if(GUI_mode == GUI_TK && !tk_ok())
	error("No DISPLAY set.\n");
#endif

    /* Detach if needed */
    if(background) {
	write_pid(0); /* Dummy to make sure we detect a stale lock
		       * before forking.
		       */
	switch(fork()) {
	  case -1:
	    error("fork() failed.\n");
	  case 0:
	    write_pid(getpid());
	    break;
	  default:
	    exit(0);
	}
    }

    /* Rediect output if needed */
    if(totallyQ || background) {
	fd=open("/dev/null", O_RDWR, 0);
        dup2(fd, STDIN_FILENO);
	dup2(fd, STDOUT_FILENO);
	dup2(fd, STDERR_FILENO);
	close(fd);
    }

    opt=default_opt;
}

void reinit_options(char *options)
{
    reset_options(&default_opt);
    str2array(options);
    av=tmp_av;
    ac=tmp_ac;
    process_options(&default_opt, OPTIONS_SETGLOBAL, 0);
    process_options(&default_opt, OPTIONS_SET, 0);
    opt=default_opt;
    if(loaded)
	process_options_str(M.options_str, &opt);
}


void reset_options(struct options *o)
{
    bzero((void *)o, sizeof(struct options));
    o->max_lines=4711;   /* As many as possible   */
    o->active_voices=-1; /* All voices active     */
    o->maxtime=0;        /* No maxtime by default */
    
    o->tempo=125;        /* PAL-tempo             */
    o->speed=6;
    
    o->speed0stop=1;     /* PT3.0 compatability         */
    o->click_removal=0;  /* Clickremoval off by default */
    o->low_note=0;       /* Let patternloader set these */
    o->high_note=0;
    
    o->loop_module=0;    /* No forced looping by default */
    o->break_loops=1;    /* Break loops  */
    o->auto_next=1;      /* Auto-next on */

    o->voice_detail=0;   /* Default to not changing it  */
    o->dir_components=0; /* Strip it all off by default */

    /* Default for -I is set in main.c */
}


void load_rcfile(void)
{
    int tmp;
    char buf[256];
    FILE *fp;

    opt=default_opt;
    if(!(fp=fopen(expand_name(MOD_RCFILE), "r")))
        return;

    if(!fgets(buf, 256, fp)) {
	fclose(fp);
	return;
    }
    fclose(fp);
    
    if(!(tmp=strlen(buf))) /* In case we don't even get '\n' */
        tmp=1;
    buf[tmp-1]=0;
    reinit_options(buf);
}


void save_rcfile(char *options)
{
    FILE *fp;
    
    if(!*options)
	unlink(expand_name(MOD_RCFILE));
    else {
	if(!(fp=fopen(expand_name(MOD_RCFILE), "w")))
	    return;
	
	if(fputs(options, fp) == EOF || fputc('\n', fp) == EOF) {
	    fclose(fp);
	    return;
	}
	fclose(fp);
    }
}

/* Will set options according to the passed string of options */

void process_options_str(char *s, struct options *o)
{
    str2array(s);
    av=tmp_av;
    ac=tmp_ac;
    process_options(o, OPTIONS_SET, 0);
}

#define ADD_NORMAL 1
#define ADD_INT    2
#define ADD_STR    3

/* Fills in the options_str for the options beginning at av[index].
 * 'how' determines how/if to set o->*.
 * Return value is the index of the next non-option. -1 means end of array.
 */

int process_options(struct options *o, int how, int index)
{
    unsigned int num=0; /* Remove warning */
    char *s, *s2, *s3, option, next;
    char set, set_global;
    int add_type;
    
    set=set_global=0;
    if(how & OPTIONS_SET)
	set=1;
    if(how & OPTIONS_SETGLOBAL)
	set_global=1;

    options_str[0]=0;
    
    while(index < ac) {
	s=av[index];
	if(*s == '-') {
	    ++s;
	    next=0;
	    while(*s && !next) {
		option=*s;
		add_type=ADD_NORMAL; /* Default */
		switch(*s) {
#if 0
		  case 'a':
		    if(set_global)
			o->auto_next=1;
		    break;
#endif
		  case 'A':
		    if(set_global)
			o->auto_next=0;
		    break;
		  case 'b':
		    if(set_global)
			o->browse=1;
		    break;
		  case 'B':
		    if(set)
			o->break_loops=0;
		    break;
		  case 'c':
		    if(!getarg(&s, &index))
			goto no_num_arg;
		    next=1;
		    num=atoi(s);
		    if(set)
			o->click_removal=num;
		    add_type=ADD_INT;
		    break;
		  case 'C':
		    if(set_global)
			o->replay_forever=1;
		    break;
		  case 'd':
		    if(set_global)
			o->dump=1;
		    break;
		  case 'D':
		    getarg(&s, &index);
		    next=1;
		    if(set_global) {
			strcpy(startup_dir, s);
		    }
		    add_type=ADD_STR;
		    break;
		  case 'f':
		    getarg(&s, &index);
		    next=1;
		    if(!strcmp(s, "mod")) {
			if(set)
			    o->format=MODFORMAT_MOD;
		    }
		    else if(!strcmp(s, "ult")) {
			if(set)
			    o->format=MODFORMAT_ULT;
		    }
		    else if(!strcmp(s, "mtm")) {
			if(set)
			    o->format=MODFORMAT_MTM;
		    }
		    else if(!strcmp(s, "s3m")) {
			if(set)
			    o->format=MODFORMAT_S3M;
		    }
		    else
			error("Unknown format '%s' specified.\n", s);
		    add_type=ADD_STR;
		    break;
		  case '?':
		  case 'h':
		    if(set_global) {
			print_helptext();
			exit(0);
		    }
		    break;
		  case 'i':
		    getarg(&s, &index);
		    next=1;
		    if(!strcmp(s, "none") || !strcmp(s, "0")) {
			if(set_global)
			    GUI_mode=GUI_NONE;
		    }
		    else if(!strcmp(s, "simple") || !strcmp(s, "1")) {
			if(set_global)
			    GUI_mode=GUI_SIMPLE;
		    }
#ifdef COMPILE_GUI_NC
		    else if(!strcmp(s, "terminal") || !strcmp(s, "2")) {
			if(set_global)
			    GUI_mode=GUI_NC;
		    }
#endif
#ifdef COMPILE_GUI_TK
		    else if(!strcmp(s, "X") || !strcmp(s, "x") ||
			    !strcmp(s, "3")) {
			if(set_global)
			    GUI_mode=GUI_TK;
		    }
#endif
		    else
			error("Interface '%s' not supported.\n", s);
		    add_type=ADD_STR;
		    break;
		  case 'I':
		    getarg(&s, &index);
		    next=1;
#ifdef COMPILE_GUI_TK
		    if(!strcmp(s, "chubby") || !strcmp(s, "1")) {
			if(set_global)
			    GUI_size=1;
		    }
		    else if(!strcmp(s, "slim") || !strcmp(s, "2")) {
			if(set_global)
			    GUI_size=2;
		    }
		    else if(!strcmp(s, "normal") || !strcmp(s, "3")) {
			if(set_global)
			    GUI_size=3;
		    }
		    else if(!strcmp(s, "verbose") || !strcmp(s, "4")) {
			if(set_global)
			    GUI_size=4;
		    }
		    else
			error("Unknown window size '%s'.\n", s);
		    add_type=ADD_STR;
#else
		    error("X interface not compiled in.\n");
#endif
		    break;
		  case 'k':
		    if(set_global)
			error("Option '-k' must be the only argument.\n");
		    break;
		  case 'l':
		    if(!getarg(&s, &index))
			goto no_num_arg;
		    next=1;
		    num=MAX(atoi(s), 1);
		    if(set_global)
			o->max_lines=num;
		    add_type=ADD_INT;
		    break;
		  case 'L':
		    if(set)
			o->loop_module=1;
		    break;
		  case 'm':
		    if(!getarg(&s, &index))
			goto no_num_arg;
		    next=1;
		    num=MAX(atoi(s), 1);
		    if(set)
			o->maxtime=num;
		    add_type=ADD_INT;
		    break;
		  case 'M':
		    if(set)
			o->mono=1;
		    break;
		  case 'n':
		    if(set_global)
			o->noscroll=1;
		    break;
		  case 'N':
		    if(set)
			o->ntsc_samples=1;
		    break;
		  case 'o':
		    if(set) {
			o->low_note=BASE_NOTE+3*12;
			o->high_note=BASE_NOTE+6*12-1;
		    }
		    break;
		  case 'O':
		    if(set) {
			o->low_note=BASE_NOTE+0*12;
			o->high_note=BASE_NOTE+NR_OCTAVES*12-1;
		    }
		    break;
		  case 'p':
		    if(!getarg(&s, &index))
			goto no_num_arg;
		    next=1;
		    num=MAX(atoi(s), 0);
		    if(set)
			o->start_pos=num;
		    add_type=ADD_INT;
		    break;
#if 0
		  case 'P':
		    if(set)
			o->ntsc_samples=0;
		    break;
#endif
		  case 'Q':
		    if(set_global)
			totallyQ=1;
		    /* Fall through to 'q' */
		  case 'q':
		    if(set_global)
			quiet=1;
		    break;
		  case 'r':
		    if(set_global)
			o->random_mode=1;
		    break;
		  case 's':
		    if(!getarg(&s, &index))
			goto no_num_arg;
		    next=1;
		    num=MAX(atoi(s), 1);
		    if(set)
			o->speed=num;
		    add_type=ADD_INT;
		    break;
		  case 'S':
		    getarg(&s, &index);
		    next=1;
		    num=strtol(s, &s2, 10);
		    if(s == s2 || *s2 != ',') {
		      wrong_arg:
			error("Argument must be in form 'x,y' where x and y "
			      "are positive integers.\n");
		    }
		    if(set)
			o->stop_songpos=num;
		    num=strtol(++s2, &s3, 10);
		    if(s2 == s3)
			goto wrong_arg;
		    if(set)
			o->stop_linepos=num;
		    add_type=ADD_STR;
		    break;
		  case 't':
		    if(!getarg(&s, &index))
			goto no_num_arg;
		    next=1;
		    num=MIN(MAX(atoi(s), 32), 255);
		    if(set)
			o->tempo=num;
		    add_type=ADD_INT;
		    break;
		  case 'T':
		    if(set)
			o->tolerant=1;
		    break;
		  case 'v':
		    if(set_global)
			o->verbose++;
		    break;
		  case 'x':
		    if(!getarg(&s, &index))
			goto no_num_arg;
		    next=1;
		    num=MAX(1, MIN(3, atoi(s)));
		    if(set)
			o->voice_detail=num;
		    add_type=ADD_INT;
		    break;
		  case 'X':
		    if(!getarg(&s, &index))
			goto no_num_arg;
		    next=1;
		    num=atoi(s);
		    if(set_global)
			o->dir_components=num;
		    add_type=ADD_INT;
		    break;
		  case 'Y':
		    getarg(&s, &index);
		    next=1;
		    if(set_global) {
			strcpy(mixer_namebuf, s);
			mixer_name=mixer_namebuf;
		    }
		    break;
		  case 'z':
		    if(set_global) {
			background=1;
			quiet=1;
		    }
		    break;
		  case 'Z':
		    if(set_global)
			o->allow_delete=1;
		    break;
		  case '0':
		    if(set)
			o->speed0stop=0;
		    break;
		  case '5':
		    if(set) {
			o->tempo=125; /* 50 Hz (125 BPM) -> 2.000/100 s */
			o->nobpm=1;
		    }
		    break;
		  case '6':
		    if(set) {
			o->tempo=150; /* 60 Hz (150 BPM) -> 1.667/100 s */
			o->nobpm=1;
		    }
		    break;
		  default:
		    error("%s: Unknown option -- %c\n", exe_name, *s);
		    break;
		  no_num_arg:
		    error("%s: option -%c requires a (positive)"
			  " numerical argument\n", exe_name, option);
		}

		switch(add_type) {
		  case ADD_NORMAL:
		    add_opt(option);
		    break;
		  case ADD_INT:
		    add_opt_int(option, num);
		    break;
		  case ADD_STR:
		    add_opt_str(option, s);
		    break;
		  default:
		    error("Unknown add_type.\n");
		}
		++s;
	    }
	    ++index;
	}
	else
	    break; /* Non-option found */
    }

    if(set && o->loop_module)    /* Loop as composer intended */
	o->break_loops=0;
    
    /* Remove the leading space */
    if(options_str[0] == ' ')
	memmove(options_str, options_str+1, strlen(options_str));
    
    if(index < ac)
	return index;
    
    return -1;
}


void str2array(char *str)
{
    int str_pos, tmp_pos;
    
    str_pos=tmp_ac=tmp_pos=0;
    
    /* Build the argument-array */
    while(str[str_pos]) {
	tmp_av[tmp_ac++]=&tmp_buf[tmp_pos];
	while(str[str_pos] && str[str_pos] != ' ' &&
	      str[str_pos] != '\t')
	    tmp_buf[tmp_pos++]=str[str_pos++];
	tmp_buf[tmp_pos++]=0;
	
	while(str[str_pos] && (str[str_pos] == ' ' || str[str_pos] == '\t'))
	    ++str_pos;
    }
    tmp_buf[tmp_pos]=0;
    tmp_av[tmp_ac]=&tmp_buf[tmp_pos];
}


void add_opt(char option)
{
    char buf[]=" -?";
    
    buf[2]=option;
    strcat(options_str, buf);
}

void add_opt_str(char option, char *str)
{
    char buf[]=" -?";
    
    buf[2]=option;
    strcat(options_str, buf);
    strcat(options_str, str);
}

void add_opt_int(char option, int num)
{
    char buf[32];

    sprintf(buf, " -%c%d", option, num);
    strcat(options_str, buf);
}

/* Sets *s to the start of the argument. Returns true if the first character
 * of the argument is a digit.
 */
int getarg(char **s, int *index)
{
    *s+=1;
    if(!**s) {
	*index=*index+1;
	if(*index >= ac)
	    error("%s: Missing argument for option "
		  "-- %c\n", exe_name, *(*s-1));
	*s=av[*index];
    }
    
    return (**s >= '0' && **s <= '9');
}
