/*
 *  util.c - Various support functions.
 *
 *  (C) 1994, 1995 Mikael Nordqvist (d91mn@efd.lth.se, mech@df.lth.se)
 */

#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/soundcard.h>
#include <sys/stat.h>
#include <dirent.h>
#include <pwd.h>

#include "mod.h"
#include "message.h"

SEQ_DECLAREBUF();

/*
 *  Wordswapping functions
 */

unsigned short SWAPSHORT(char *a)
{
    return (((*(unsigned short *)a&0xff)<<8)|((*(unsigned short *)a>>8)&0xff));
}

unsigned long SWAPLONG(char *a)
{
    return (SWAPSHORT(a)<<16 | SWAPSHORT(a+2));
}

/*
 *  Removes nonprintable characters and trailing spaces from a string.
 */

void fix_string(char *s)
{
    char *a;

    a=s;
    while(*s) {
	if(*s >= 32 && *s <=126)
	    s++;
	else
	    *s++=' ';
    }

    for(--s; s >=a; --s)  {
	if(*s == ' ')
	    *s=0;
	else
	    break;
    }
}

/* Fills out a string with spaces, or truncates it if its too long. */

char *fill_string(char *s, int len)
{
    static char buf[128];
    int i;

    strcpy(buf, s);

    i=strlen(s);
    if(i > len)
	i=len;

    for(; i < len; ++i)
	buf[i]=' ';
    
    buf[i]=0;
    return buf;
}


/* Returns a pointer to the filename with all but 'nr' directories from
 * the path removed.
 */

char *path_basename(char *s, int nr)
{
    int i;
    
    i=strlen(s)-1;
    for(++nr; nr; --nr) {
	for(; i >=0 && s[i] != '/'; --i)
	    ;
	if(nr > 1 && --i < 0) {
	    i=-1;
	    break;
	}
    }
    if(i == 1 && *s == '/')
	i=-1;
    
    return &s[++i];
}

/* Truncates a filename */

char *trunc_filename(char *n, int maxlen)
{
    static char buf[80];
    int i, len, strip;
    char *s;
    
    len=PATH_MAX;
    strip=42; /* Strip 42 directory-components for a start */
    do {
	s=path_basename(n, strip--);
	i=strlen(s);
	len=i;
    }
    while(len > maxlen && strip);
    
    if(strlen(s) >= strlen(n)-1) {
	if(strlen(s)+1 == strlen(n))
	    s=n;
	strcpy(buf, s);
    }
    else {
	strcpy(buf, ".../");
	strcat(buf, s);
    }
    return buf;
}


/* Eats (non)whitespace starting at buf[*p].
 * dir = +1, search forward
 * dir = -1, search backward
 */

void eat_ws(int *pp, char *buf, int dir)
{
    int p=*pp;
    do {
	if(!(buf[p] == ' ' || buf[p] == '\t'))
	    break;
	p+=dir;
    } while(p > 0 && buf[p]);
    *pp=p;
}

void eat_nonws(int *pp, char *buf, int dir)
{
    int p=*pp;
    do {
	if(buf[p] == ' ' || buf[p] == '\t')
	    break;
	p+=dir;
    } while(p > 0 && buf[p]);
    *pp=p;
}


/*
 *  This is seqbuf_dump() used by the SEQ_DUMPBUF macro.
 */

void seqbuf_dump(void)
{
    int thisone=0; /* No need to initialize, but it removes a warning */
    fd_set wset;

    while(_seqbufptr) {
	FD_ZERO(&wset);
	FD_SET(seqfd, &wset);
	if(select(seqfd+1, 0, &wset, 0, 0) >= 0) {
	    if((thisone=write(seqfd, _seqbuf, _seqbufptr)) == -1)
		error("Unable to write(sequencer)");

	    if(thisone < _seqbufptr) {
		memmove(_seqbuf, &(_seqbuf[thisone]), _seqbufptr-thisone);
		_seqbufptr-=thisone;
	    }
	    else
		_seqbufptr=0;
	}
	else {
	    if(errno != EINTR)
		error("select() in seqbuf_dump failed (errno=%d)\n", errno);
	}
    }
}

/*
 *  Takes track-data and returns the corresponding track-index.
 *  This is where identical tracks are found.
 */

int get_track_idx(struct event *e)
{
    int i;
    
    for(i=0; i < M.nr_tracks; ++i)
	if(!bcmp(e, M.tracks[i], sizeof(struct event)*64))
	    return i;
    
    M.tracks[M.nr_tracks]=(struct event *)malloc(sizeof(struct event)*64);
    memcpy((void *)M.tracks[M.nr_tracks], (void *)e, sizeof(struct event)*64);
    return M.nr_tracks++;
}


/* Allows reading of a small amounts of bytes from a file, n <= 10. Buffers
 * data to minimize number of read()-calls.
 * n = 0   Init.
 * n = -1  Cleanup (only needed if you need the read-ahead flushed).
 * The function returns true if all bytes could be read, false otherwise.
 */

int get_bytes(int fd, char *dest, int n)
{
    static char buf[1024];
    static int bufpos, lastbufpos, nomore;
    int thisone;

    if(n == -1) {
	if(lseek(fd, -(lastbufpos-bufpos+1), SEEK_CUR) == -1)
	    error("Seek() failed.\n");
	return 1;
    }
    
    if(!n) {
	bufpos=0;
	lastbufpos=read(fd, buf, 1024);
	if(!lastbufpos || lastbufpos == -1) {
	    return 0;
	}
	else {
	    if(lastbufpos != 1024)
		nomore=1;
	    else
		nomore=0;
	    --lastbufpos;
	    return 1;
	}
    }

    if(bufpos+n-1 <= lastbufpos) {
	memcpy(dest, &buf[bufpos], n);
	bufpos+=n;
    }
    else {
	return 0;
    }

    if(bufpos > 1010 && !nomore) {
	memmove(buf, &buf[bufpos], 1024-bufpos);
	if((thisone=read(fd, &buf[1024-bufpos], bufpos)) == 0 ||
	   thisone == -1 || thisone != bufpos) {
	    nomore=1;
	    if(thisone <= 0)
		thisone=0;
	    lastbufpos=1024-bufpos+thisone-1;
	}
	bufpos=0;
    }
    return 1;
}


/*
 * Reads a string from a filedescriptor
 */

void safe_read(int fd, struct mod_message *buf, int len)
{
    int sofar, thisone;
    
    sofar=0;
    do {
	thisone=read(fd, ((char *)buf)+sofar, len-sofar);
	if(thisone >= 0) {
	    sofar+=thisone;
	    if(!thisone)
		debug("safe_read(): 0\n");
	}
	else {
	    if(errno != EAGAIN) {
		debug("Error reading pipe %d "
		      "(seq=%d) (ack[0]=%d) (p2m_s[0]=%d) (m2p[0]=%d) "
		      "(errno=%d)\n",
		      fd,
		      seqfd, ack_pipe[0], player2main_synced[0],
		      main2player[0], errno);
	    }
	    else
		debug("safe_read() returned EAGAIN, retrying...\n");
	}
    }
    while(sofar < len);
}


/*
 * Writes a string to a filedescriptor
 */

void safe_write(int fd, struct mod_message *buf, int len)
{
    int sofar, thisone;
    
    sofar=0;
    do {
	thisone=write(fd, ((char *)buf)+sofar, len-sofar);
	if(thisone >= 0) {
	    sofar+=thisone;
	    if(!thisone)
		debug("safe_write(): 0\n");
	}
	else {
	    if(errno != EAGAIN)
		error("Error writing pipe (%d)\n", errno);
	    error("safe_write(fd=%d len=%d) would block...pid_player=%d\n",
		  fd, len, pid_player);
	}
    }
    while(sofar < len);
}

/* Escapes characters for the shell (or tcl) */

char *escape_name(char *s)
{
    static char buf[PATH_MAX+1];
    char *d;
    
    d=buf;
    while(*s) {
	if(!((*s >= 'a' && *s <= 'z') || (*s >= 'A' && *s <= 'Z') ||
	     (*s >= '0' && *s <= '9') || *s == '.' || *s == '/' || *s == '-'
	     || *s == '_'))
	    *d++='\\';
	*d++=*s++;
    }
    *d=0;
    return buf;
}

/* As above, but if first char is ~ that char won't be escaped */

char *escape_name_tilde(char *s)
{
    s=escape_name(s);
    if(!strncmp("\\~", s, 2))
	s++;
    return s;
}



/* Returns < 0 if we got SIG_QUIT otherwise exitstatus of command */

int my_system(char *a)
{
    int p, status;

    switch(p=fork()) {
      case -1:
	error("Fork failed in my_system.");
	break;
      case 0:
	exit(system(a));
	break;
      default:
	while(1) {
	    if(waitpid(p, &status, 0) >= 0) {
		if(WIFEXITED(status)) {
		    if(quit == QUIT_SIGNAL)
			return -1;
		    else
			return WEXITSTATUS(status);
		}
		else {
		    debug("Huh??!! system()-problems\n");
		    return 1;
		}
	    }
	    else {
		if(errno != EINTR)
		    debug("Huh??!! waitpid() returned errno=%d.\n", errno);
	    }
	}
    }
    error("Never reached");
    return 0; /* Remove warning */
}

/* Removes a directory (recursively) located in the _current directory_ */

void my_removedir(char *name)
{
    DIR *d;
    struct dirent *de;
    struct stat s;

    if(!chdir(name)) {
	if((d=opendir("."))) {
	    while((de=readdir(d))) {
		if(strcmp(de->d_name, ".") && strcmp(de->d_name, "..") &&
		   strcmp(de->d_name, "mod.pid")) {
		    if(stat(de->d_name, &s))
			continue;
		    chmod(de->d_name, 0777); /* a+rwx */
		    if(S_ISDIR(s.st_mode)) {
			my_removedir(de->d_name);
		    }
		    else
			unlink(de->d_name);
		}
	    }
	    chmod("mod.pid", 0777); /* a+rwx */
	    unlink("mod.pid"); /* Remove this one last if it exists.
				* Can't remember WHY I wait with that one...
				*/
	    closedir(d);        
	}
	chdir("..");
    }
    rmdir(name); /* May or may not work depending on owner of dir */
}

/* Expands ~ in a filename (if it's present) */

char *expand_name(char *s)
{
    static char buf[PATH_MAX+1];
    char *a, tmp;
    struct passwd *pwe;
    
    if(*s != '~')
	return s;
    for(a=s+1; *a && *a != '/'; ++a)
	;
    if(a == s+1) { /* Users homedirectory */
	pwe=getpwuid(getuid());
    }
    else {
	tmp=*a;
	*a=0;
	pwe=getpwnam(s+1);
	*a=tmp;
    }
    if(!pwe)
	return s;
    strcpy(buf, pwe->pw_dir);
    strcat(buf, a);
    return buf;
}
